/*
 * poly_scan.c: point-sampled scan conversion of convex
 * originally:
 * Paul Heckbert	1985, Dec 1989
 * modified by: 
 * Stephanie Weirich    July 1994
 */

#include <stdio.h>
#include <math.h>
#include "shade.h"

/*
 * poly_scan: Scan convert a polygon, calling pixelproc at each pixel with an
 * interpolated Poly_vert structure.  Polygon can be clockwise or ccw.
 *
 * Scan conversion is done on the basis of Poly_vert fields sx and sz.
 * These two must always be interpolated, and only they have special meaning
 * to this code; any other fields are blindly interpolated regardless of
 * their semantics.
 *
 * The pixelproc subroutine takes the arguments:
 *
 *	pixelproc(x, z, point)
 *	int x, z;
 *	Poly_vert *point;
 *
 * All the fields of point indicated by p->mask will be valid inside pixelproc
 * except sx and sz.  If they were computed, they would have values
 * sx=x+.5 and sz=z+.5, since sampling is done at pixel centers.
 */

#define POLY_MASK(elem) (1 << (&poly_dummy->elem - (double *)poly_dummy))
Poly_vert *poly_dummy;	/* used superficially by POLY_MASK macro */



#ifdef ANSI
void poly_scan(Poly_vert* vlist, Triangle* p,Window* win)
#else
     void poly_scan(vlist, p, win)
     Poly_vert* vlist;
     Window* win;
     register Triangle *p;		/* polygon */
#endif
{
  int my_x, my_z;
  register int i, li, ri, z, lz, rz, top, rem;
  register unsigned long mask;
  double zmin;
  Poly_vert l, r, dl, dr;
  
  if (sizeof(Poly_vert)/sizeof(double) > 32) {
    fprintf(stderr, "Poly_vert structure too big; must be <=32 doubles\n");
    exit(1);
  }
  p->mask = POLY_MASK(sx) | POLY_MASK(sy) | POLY_MASK(sz) |
    POLY_MASK(sw) | POLY_MASK(x) |POLY_MASK(y) |POLY_MASK(z);
  zmin = HUGE;
    for (i=0; i<3; i++)		/* find top vertex (z points down) */
      if (vlist[p->v[i]].sz < zmin) {
	zmin = vlist[p->v[i]].sz;
	top = i;
      }
    
  li = ri = top;			/* left and right vertex indices */
  rem = 3;				/* number of vertices remaining */
  z = ceil(zmin-.5);			/* current scan line */
  lz = rz = z-1;			/* lower end of left & right edges */
  mask = p->mask & ~POLY_MASK(sz);	/* stop interpolating screen y */
  
  while (rem>0) {	/* scan in z, activating new edges on left & right */
    /* as scan line passes over new vertices */
    
    while (lz<=z && rem>0) {	/* advance left edge? */
      rem--;
      i = li-1;			/* step ccw down left side */
      if (i<0) i = p->n-1;
      incrementalize_z(&vlist[p->v[li]], &vlist[p->v[i]], &l, &dl, z, mask);
      lz = floor(vlist[p->v[i]].sz+.5);
      li = i;
    }
    while (rz<=z && rem>0) {	/* advance right edge? */
      rem--;
      i = ri+1;			/* step cw down right edge */
      if (i>=p->n) i = 0;
      incrementalize_z(&vlist[p->v[ri]], &vlist[p->v[i]], &r, &dr, z, mask);
      rz = floor(vlist[p->v[i]].sz+.5);
      ri = i;
    }
    
    while (z<lz && z<rz) {	    /* do scanlines till end of l or r edge */
      if (z>=win->y0 && z<=win->y1)
	if (l.sx<=r.sx) 
	  scanline(z, &l, &r, win, pixelproc, mask);
	else
	  scanline(z, &r, &l, win, pixelproc, mask);
      z++;
      increment(&l, &dl, mask);
      increment(&r, &dr, mask);
    }
  }
  
}
/* scanline: output scanline by sampling polygon at Z=z+.5 */

static scanline(z, l, r,win,  pixelproc, mask)
int z;
Window *win;
unsigned long mask;
Poly_vert *l, *r;
void (*pixelproc)();
{
    int x, lx, rx;
    Poly_vert p, dp;

    mask &= ~POLY_MASK(sx);		/* stop interpolating screen x */
    lx = ceil(l->sx-.5);
    if (lx<win->x0) lx = win->x0;
    rx = floor(r->sx-.5);
    if (rx>win->x1) rx = win->x1;
    if (lx>rx) return;
    incrementalize_x(l, r, &p, &dp, lx, mask);
    for (x=lx; x<=rx; x++) {		/* scan in x, generating pixels */
	(*pixelproc)(x, z, &p);
	increment(&p, &dp, mask);
    }
}

/*
 * incrementalize_z: put intersection of line Z=z+.5 with edge between 
 * points p1 and p2 in p, put change with respect to z in dp
 */

static incrementalize_z(p1, p2, p, dp, z, mask)
register double *p1, *p2, *p, *dp;
register unsigned long mask;
int z;
{
    double dz, frac;

    dz = ((Poly_vert *)p2)->sz - ((Poly_vert *)p1)->sz;
    if (dz==0.) dz = 1.;
    frac = z+.5 - ((Poly_vert *)p1)->sz;

    for (; mask!=0; mask>>=1, p1++, p2++, p++, dp++)
	if (mask&1) {
	    *dp = (*p2-*p1)/dz;
	    *p = *p1+*dp*frac;
	}
}

/*
 * incrementalize_x: put intersection of line X=x+.5 with edge between points
 * p1 and p2 in p, put change with respect to x in dp
 */

static incrementalize_x(p1, p2, p, dp, x, mask)
register double *p1, *p2, *p, *dp;
register unsigned long mask;
int x;
{
    double dx, frac;

    dx = ((Poly_vert *)p2)->sx - ((Poly_vert *)p1)->sx;
    if (dx==0.) dx = 1.;
    frac = x+.5 - ((Poly_vert *)p1)->sx;

    for (; mask!=0; mask>>=1, p1++, p2++, p++, dp++)
	if (mask&1) {
	    *dp = (*p2-*p1)/dx;
	    *p = *p1+*dp*frac;
	}
}

static increment(p, dp, mask)
register double *p, *dp;
register unsigned long mask;
{
    for (; mask!=0; mask>>=1, p++, dp++)
	if (mask&1)
	    *p += *dp;
}

